"""
Nagios Plugin resource(s).
"""
from logging import getLogger

from nagiosplugin import (
    CheckError,
    Metric,
    Resource,
)


COUNT = "count"
MEAN = "mean"
VALUES = "values"


class Measurements(Resource):
    """
    Count and mean metrics for a simple InfluxDB measurement query.
    """
    def __init__(self,
                 query,
                 client):
        self.query = query
        self.client = client
        self.logger = getLogger('nagiosplugin')

    def get_results(self):
        """
        Fetch results from InfluxDB.
        """
        try:
            return self.query.get_results(self.client)
        except Exception as error:
            self.logger.info("Failed to query InfluxDB: {}".format(
                error,
            ))
            raise CheckError(error)

    def probe(self):
        """
        Query InfluxDB; yield the count and mean of the measurements.
        """
        def get_value(result):
            # measurement results should have a 'value' field
            if "value" in result:
                return result["value"]
            # otherwise, 'name' is a good guess
            return result["name"]

        values = [
            get_value(result) for result in self.get_results()
        ]

        count = len(values)
        yield Metric(COUNT, count, context=COUNT)

        try:
            total = float(sum(values))
            mean = 0 if count == 0 else total / count
            yield Metric(MEAN, mean, context=MEAN)
        except TypeError:
            # non numeric queries won't have a mean
            pass

        # null context; values are not validated individually
        yield Metric(VALUES, values, context="null")



"""
InfluxDB query building.
"""
from abc import ABCMeta, abstractmethod
from logging import getLogger

from six import add_metaclass


DEFAULT_FIELDS = [
    "time",
    "value",
]


def maybe_quote(value):
    """
    Quote a value for InfluxDB if necessary.
    """
    if value[0] == "'" and value[-1] == "'":
        return value
    return "'{}'".format(value)


def kv_condition(key, value):
    """
    Generate a key value equality condition.

    Ensure that the value is properly quoted.
    """
    return "{} = {}".format(key, maybe_quote(value))


def age_condition(age):
    return "time > now() - {}".format(age)


def host_condition(hostname):
    return kv_condition("host", hostname)


def kv_conditions(conditions):
    return [
        kv_condition(*condition.split("=", 1))
        for condition in conditions or []
    ]


class Query(object):
    """
    Query wrapper.
    """
    def __init__(self, query, measurements=None):
        self.query = query
        self.measurements = measurements or []
        self.logger = getLogger('nagiosplugin')

    def __str__(self):
        return self.query

    def get_results(self, client):
        self.logger.debug("Querying InfluxDB at {}:{} with query: \"{}\"".format(
            client._host,
            client._port,
            self.query,
        ))
        results = client.query(self.query)
        self.logger.info("Received result set: {}".format(results))
        return list(results.get_points())


@add_metaclass(ABCMeta)
class QueryBuilder(object):
    """
    Abstract callable that builds queries.
    """
    @abstractmethod
    def __call__(self):
        pass


class ExplicitQueryBuilder(QueryBuilder):
    """
    Return the provided query.
    """
    def __init__(self, query):
        self.query = query

    def __call__(self):
        return Query(
            query=self.query,
        )


class SingleMeasurementQueryBuilder(QueryBuilder):
    """
    Build a simple InfluxDB query of the form:

        SELECT <fields> FROM <measurement> WHERE <conditions>
    """
    def __init__(self,
                 fields=None,
                 measurement=None,
                 conditions=None):
        self.fields = fields or ["time", "value"]
        self.measurement = measurement
        self.conditions = conditions or []

    def __call__(self):
        return Query(
            query="SELECT {} FROM {} WHERE {}".format(
                ", ".join(self.fields),
                self.measurement,
                " AND ".join(self.conditions),
            ),
            measurements=[
                self.measurement,
            ],
        )

    @classmethod
    def for_hostname_and_age(cls, measurement, age, hostname, where):
        return cls(
            fields=DEFAULT_FIELDS,
            measurement=measurement,
            conditions=[
                age_condition(age),
                host_condition(hostname),
            ] + kv_conditions(where),
        )



"""
Command line entry point.
"""
from click import argument, group, option
from influxdb import InfluxDBClient
from nagiosplugin import (
    Check,
    guarded,
    ScalarContext,
)

from influxdbnagiosplugin.query import ExplicitQueryBuilder, SingleMeasurementQueryBuilder
from influxdbnagiosplugin.resources import COUNT, MEAN, Measurements
from influxdbnagiosplugin.summaries import MeasurementValuesSummary


@group(chain=True)
@option("-v", "--verbose", count=True)
@option("--hostname", default="localhost", help="InfluxDB hostname")
@option("--port", default=8086, help="InfluxDB port")
@option("--username", default="influxdb", help="InfluxDB usernanme")
@option("--password", default="secret", help="InfluxDB password")
@option("--database", default="telegraf", help="InfluxDB database name")
@option("--count-error-range", default="1:", help="Range of measurement counts that are NOT considered an error")  # noqa
@option("--count-warning-range", default="2:", help="Range of measurement counts that are NOT considered a warning")  # noqa
@option("--mean-error-range", default="", help="Range of measurement means that are NOT considered an error")  # noqa
@option("--mean-warning-range", default="", help="Range of measurement counts that are NOT considered a warning")  # noqa
def main(**kwargs):
    """
    Command line entry point. Defines common arguments.
    """
    pass


@main.resultcallback()
@guarded
def check(processors, **args):
    """
    Invoke the InfluxDB check using the Nagios plugin framework.

    Reads the query from the processors chain.
    """
    query = processors[0]()

    client = InfluxDBClient(
        host=args["hostname"],
        port=args["port"],
        username=args["username"],
        password=args["password"],
        database=args["database"],
    )

    check = Check(
        Measurements(
            query=query,
            client=client,
        ),
        ScalarContext(COUNT, args["count_warning_range"], args["count_error_range"]),
        ScalarContext(MEAN, args["mean_warning_range"], args["mean_error_range"]),
        MeasurementValuesSummary(
            query=query,
        )
    )
    check.main(args["verbose"])


@main.command()
@argument("query")
def query(query):
    """
    Run an explicit query.
    """
    return ExplicitQueryBuilder(query)


@main.command()
@argument("measurement")
@argument("hostname")
@option("--age", default="30s")
@option("--where", multiple=True, help="Extra where conditions to include.")
def single(measurement, hostname, age, where):
    """
    Run a query for a single measurement.
    """
    return SingleMeasurementQueryBuilder.for_hostname_and_age(
        measurement=measurement,
        hostname=hostname,
        age=age,
        where=where,
    )
